Package: airr
Type: Package
Version: 2.0.0
Date: 2026-06-05
Authors@R: c(person("Jason", "Vander Heiden", role=c("aut", "cre"), 
                    email="jason.vanderheiden@gmail.com"),
             person("Susanna", "Marquez", role=c("aut"), 
                    email="susanna.marquez@yale.edu"),
             person("Scott", "Christley", role=c("aut"), 
                    email="Scott.Christley@UTSouthwestern.edu"),
             person("Katharina", "Imkeller", role=c("aut"), 
                    email="imkeller@embl.de"),
             person("Ulrik", " Stervbo", role=c("aut"),
                    email="ulrik.stervbo@elisabethgruppe.de"),
             person("Gisela", "Gabernet", role=c("aut"),
                    email="gisela.gabernet@yale.edu"),
             person("AIRR Community", role=c("cph")))
Title: AIRR Data Representation Reference Library
Description: Schema definitions and read, write and validation tools for data 
    formatted in accordance with the AIRR Data Representation schemas defined 
    by the AIRR Community <https://docs.airr-community.org>.
License: CC BY 4.0
URL: https://docs.airr-community.org
BugReports: https://github.com/airr-community/airr-standards/issues
BuildVignettes: true
VignetteBuilder: knitr
Encoding: UTF-8
Depends: R (>= 3.1.2)
Imports: jsonlite, methods, readr, stats, stringi, tools, yaml
Suggests: knitr, rmarkdown, tibble, testthat
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-06-06 07:38:49 UTC; vandej27
Author: Jason Vander Heiden [aut, cre],
  Susanna Marquez [aut],
  Scott Christley [aut],
  Katharina Imkeller [aut],
  Ulrik Stervbo [aut],
  Gisela Gabernet [aut],
  AIRR Community [cph]
Maintainer: Jason Vander Heiden <jason.vanderheiden@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-06 08:10:02 UTC
Built: R 4.5.2; ; 2026-06-06 08:46:07 UTC; unix

Package: airr
Type: Package
Version: 2.0.0
Date: 2026-06-05
Authors@R: c(person("Jason", "Vander Heiden", role=c("aut", "cre"), 
                    email="jason.vanderheiden@gmail.com"),
             person("Susanna", "Marquez", role=c("aut"), 
                    email="susanna.marquez@yale.edu"),
             person("Scott", "Christley", role=c("aut"), 
                    email="Scott.Christley@UTSouthwestern.edu"),
             person("Katharina", "Imkeller", role=c("aut"), 
                    email="imkeller@embl.de"),
             person("Ulrik", " Stervbo", role=c("aut"),
                    email="ulrik.stervbo@elisabethgruppe.de"),
             person("Gisela", "Gabernet", role=c("aut"),
                    email="gisela.gabernet@yale.edu"),
             person("AIRR Community", role=c("cph")))
Title: AIRR Data Representation Reference Library
Description: Schema definitions and read, write and validation tools for data 
    formatted in accordance with the AIRR Data Representation schemas defined 
    by the AIRR Community <https://docs.airr-community.org>.
License: CC BY 4.0
URL: https://docs.airr-community.org
BugReports: https://github.com/airr-community/airr-standards/issues
BuildVignettes: true
VignetteBuilder: knitr
Encoding: UTF-8
Depends: R (>= 3.1.2)
Imports: jsonlite, methods, readr, stats, stringi, tools, yaml
Suggests: knitr, rmarkdown, tibble, testthat
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-06-06 07:38:49 UTC; vandej27
Author: Jason Vander Heiden [aut, cre],
  Susanna Marquez [aut],
  Scott Christley [aut],
  Katharina Imkeller [aut],
  Ulrik Stervbo [aut],
  Gisela Gabernet [aut],
  AIRR Community [cph]
Maintainer: Jason Vander Heiden <jason.vanderheiden@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-06 08:10:02 UTC
Built: R 4.6.0; ; 2026-06-06 09:40:46 UTC; unix

# Generated by roxygen2: do not edit by hand

export(AIRRSchema)
export(DataFileSchema)
export(GenotypeSetSchema)
export(GermlineSetSchema)
export(InfoSchema)
export(RearrangementSchema)
export(RepertoireSchema)
export(load_schema)
export(read_airr)
export(read_rearrangement)
export(read_tabular)
export(validate_airr)
export(validate_rearrangement)
export(validate_tabular)
export(write_airr)
export(write_rearrangement)
export(write_tabular)
exportClasses(Schema)
exportMethods("$")
exportMethods("[")
exportMethods(names)
import(methods)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(readr,col_character)
importFrom(readr,col_double)
importFrom(readr,col_integer)
importFrom(readr,col_logical)
importFrom(readr,cols)
importFrom(readr,read_tsv)
importFrom(readr,write_tsv)
importFrom(stats,setNames)
importFrom(stringi,stri_trim)
importFrom(tools,file_ext)
importFrom(yaml,read_yaml)
importFrom(yaml,write_yaml)

AIRR Standards Reference Library
-------------------------------------------------------------------------------

`airr` is an R package for working with data formatted according to the AIRR 
Standards schemas. It includes the full set of schema definitions
along with simple functions for read, write and validation.

## -----------------------------------------------------------------------------
# Imports
library(airr)
library(tibble)

# Read Rearrangement example file
f1 <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")
rearrangement <- read_rearrangement(f1)
glimpse(rearrangement)

## -----------------------------------------------------------------------------
# Read Repertoire example file
f2 <- system.file("extdata", "repertoire-example.yaml", package="airr")
repertoire <- read_airr(f2)
glimpse(repertoire)

# Read GermlineSet example file
f3 <- system.file("extdata", "germline-example.json", package="airr")
germline <- read_airr(f3)
glimpse(germline)

## -----------------------------------------------------------------------------
x1 <- file.path(tempdir(), "airr_out.tsv")
write_rearrangement(rearrangement, x1)

## -----------------------------------------------------------------------------
x2 <- file.path(tempdir(), "airr_repertoire_out.yaml")
write_airr(repertoire, x2, format="yaml")

x3 <- file.path(tempdir(), "airr_germline_out.json")
write_airr(germline, x3, format="json")

## -----------------------------------------------------------------------------
# Validate Rearrangement data.frame
validate_rearrangement(rearrangement)

# Validate an AIRR Data Model
validate_airr(repertoire)

# Validate AIRR Data Model records individual 
validate_airr(germline, each=TRUE)

#### Classes ####

#' S4 class defining an AIRR standard schema
#' 
#' \code{Schema} defines a common data structure for AIRR Data Representation 
#' standards.
#' 
#' @slot    definition  name of the schema definition.
#' @slot    required    \code{character} vector of required fields.
#' @slot    optional    \code{character} vector of non-required fields.
#' @slot    properties  \code{list} of field definitions.
#' @slot    info        \code{list} schema information.
#'                  
#' @details
#' The following predefined Schema objects are defined:
#'
#' @format
#' A \code{Schema} object.
#' 
#' @seealso
#' See \link{load_schema} for loading a \code{Schema} from the definition set.
#'
#' @name         Schema-class
#' @rdname       Schema-class
#' @aliases      Schema
#' @exportClass  Schema
setClass("Schema", 
         slots=c(definition="character",
                 required="character",
                 optional="character",
                 properties="list",
                 info="list"))


#### Methods #####

#' @param    x    \code{Schema} object.
#' 
#' @rdname   Schema-class
#' @aliases  Schema-method
#' @export
setMethod("names",
          signature(x="Schema"),
          function (x) { names(x@properties) })

#' @param    i    field name.
#' 
#' @rdname   Schema-class
#' @aliases  Schema-method
#' @export
setMethod("[",
          signature(x="Schema", i="character"),
          function (x, i) { x@properties[[i]] })

#' @param    name  field name.
#' 
#' @rdname   Schema-class
#' @aliases  Schema-method
#' @export
setMethod("$",
          signature(x="Schema"),
          function (x, name) { x@properties[[name]] })


#### Schema I/O ####



# Function to extract entries from individual fields
extract_field_content <- function(properties, field) {
    
    types <- c("string"="character", "boolean"="logical", "integer"="integer", "number"="numeric", "array"="array", "object"="object")
    
    # if there is a simple reference to another AIRR schema elements, call the reference entries
    # example: disease_diagnosis entry in Diagnosis schema 
    if(!is.null(properties[[field]]$`$ref`)) {
        # name of the AIRR scheme element to refer to
        ref_element <- properties[[field]]$`$ref`
        # remove #/
        ref_element <- substr(ref_element, 3, nchar(ref_element))
        ref_schema <- load_schema(ref_element)
        # overwrite the property of the field with the Schema it is referencing to
        properties[[field]][["ref"]] <- ref_schema
    }
    
    # if there is an array type format that has ref as entries
    # example: diagnosis entry in Subject schema
    items <- unlist(properties[[field]][["items"]])
    if(!is.null(items)) {
        # check if these are references to other schemes
        if(all(grepl("ref", names(items)))) {
            ref_schemes <- items
            ref_element <- substr(ref_schemes, 3, nchar(ref_schemes))
            # store as a list of schemes
            ref_scheme_list <- sapply(ref_element, load_schema)
            properties[[field]][["ref"]] <- ref_scheme_list
        }
    }
    
    # if there is an object type, the properties may cover several reference schemes 
    # example: germline entry in Subject schema
    if(!is.null(properties[[field]][["type"]]) && 
       properties[[field]][["type"]] == "object") {
        # this should be normally defined in airr-schema.yaml as a reference scheme with optional properties
        # as a workaround we create a helper Schema without required entries
        required_helper <- character(0)
        info_helper <- list(0)
        optional_helper <- names(properties[[field]][["properties"]])
        properties_helper <- properties[[field]][["properties"]]
        
        for (f in optional_helper) {
            # if there is a reference to another AIRR schema elements, call the reference entries
            properties_helper <- extract_field_content(properties_helper, f)
        }
        
        helper_schema <- new("Schema", 
                   required=required_helper, 
                   optional=optional_helper, 
                   properties=properties_helper, 
                   info=info_helper)

        properties[[field]][["ref"]] <- helper_schema
    }
    
    x <- properties[[field]][["type"]]
    y <- properties[[field]][["description"]]
    
    # make sure that NULL types also remain NULL
    if (!is.null(x)) {properties[[field]][["type"]] <- unname(types[x])}
    properties[[field]][["description"]] <- stri_trim(y) 
    
    return(properties)
}


#' Load a schema definition
#' 
#' \code{load_schema} loads an AIRR object definition from the internal
#' definition set.
#'
#' @param    definition   name of the schema definition.
#' 
#' @return   A \link{Schema} object for the \code{definition}.
#' 
#' @details
#' Valid definitions include:
#' \itemize{
#'   \item   \code{"Rearrangement"}
#'   \item   \code{"Alignment"}
#'   \item   \code{"Repertoire"}
#'   \item   \code{"Study"}
#'   \item   \code{"Subject"}
#'   \item   \code{"Diagnosis"}
#'   \item   \code{"Sample"}
#'   \item   \code{"SampleProcessing"}
#'   \item   \code{"DataProcessing"}
#'   \item   \code{"GermlineSet"}
#'   \item   \code{"GenotypeSet"}
#' }
#'    
#' @seealso  See \link{Schema} for the return object.
#' 
#' @examples
#' # Load the Rearrangement definition
#' schema <- load_schema("Rearrangement")
#' 
#' # Load the Repertoire definition
#' schema <- load_schema("Repertoire")
#' 
#' @export
load_schema <- function(definition) {
    # Load schema from yaml file
    spec_file <- system.file("extdata", "airr-schema.yaml", package="airr")
    spec_list <- yaml::read_yaml(spec_file)
    
    # Load definition
    definition_list <- spec_list[[definition]]
    definition_list[["discriminator"]] <- definition_list[["type"]] <- NULL

    # schema info
    info <- spec_list[["Info"]]

    # Define member attributes
    if (!is.null(definition_list[["properties"]])) {
        fields <- names(definition_list[["properties"]])
        properties <- definition_list[["properties"]]
        required <- if ("required" %in% names(definition_list)) { definition_list[["required"]] } else { character() }
    } else if (!is.null(definition_list[["allOf"]])) {
        properties <- NULL
        required <- NULL
        for (s in definition_list[["allOf"]]) {
            # this logic should probably be in extract_field_content above, but it expects a field name
            if (!is.null(s[["type"]]) && s[["type"]] == "object") {
                properties <- c(properties, s[["properties"]])
            } else if (!is.null(s$`$ref`)) {
                # name of the AIRR scheme element to refer to
                ref_element <- s$`$ref`
                # remove #/
                ref_element <- substr(ref_element, 3, nchar(ref_element))
                ref_schema <- load_schema(ref_element)
                # add the properties for the Schema it is referencing to
                properties <- c(properties, ref_schema@properties)
                required <- c(required, ref_schema@required)
            }
        }
        fields <- names(properties)
    } else {
        # TODO: this is an error
    }

    optional <- setdiff(fields, required)
    
    for (f in fields) {
        # if there is a reference to another AIRR schema elements, call the reference entries
        properties <- extract_field_content(properties, f)
    }
    
    # NOTE: This bit was moved up, as it applies to any object without "required"
    # for ontology, required in NULL and this is not type character
    # this leads to a problem returning "none" in the validation function.
    # to be taken care of when array type recursion implemented
    # if (is.null(required)) { required <- character(0) }
    
    schema_object <- new("Schema", definition=definition, required=required, optional=optional, 
                        properties=properties, info=info)
    return(schema_object)
}


#### Data ####

#' Example AIRR data
#'
#' Example data files compliant with the the AIRR Data Representation standards.
#'
#' @format
#' \itemize{
#'   \item  \code{extdata/rearrangement-example.tsv.gz}: Rearrangement TSV file.
#'   \item  \code{extdata/repertoire-example.yaml}: Repertoire YAML file.
#'   \item  \code{extdata/germline-example.json}: GermlineSet and GenotypeSet JSON file.
#' }
#' 
#' @examples
#' # Load Rearrangement example
#' file <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")
#' rearrangement <- read_rearrangement(file)
#' 
#' # Load Repertoire example
#' file <- system.file("extdata", "repertoire-example.yaml", package="airr")
#' repertoire <- read_airr(file)
#' 
#' # Load GermlineSet and GenotypeSet examples
#' file <- system.file("extdata", "germline-example.json", package="airr")
#' germline <- read_airr(file)
#' 
#' @name ExampleData
NULL

#' @details   \code{InfoSchema}: AIRR Info \code{Schema}.
#' @rdname    Schema-class
#' @export
InfoSchema <- load_schema("InfoObject")

#' @details   \code{DataFileSchema}: AIRR DataFile \code{Schema}.
#' @rdname    Schema-class
#' @export
DataFileSchema <- load_schema("DataFile")

#' @details   \code{AlignmentSchema}: AIRR Alignment \code{Schema}.
#' @rdname    Schema-class
#' @export
AlignmentSchema <- load_schema("Alignment")

#' @details   \code{RearrangementSchema}: AIRR Rearrangement \code{Schema}.
#' @rdname    Schema-class
#' @export
RearrangementSchema <- load_schema("Rearrangement")

#' @details   \code{RepertoireSchema}: AIRR Repertoire \code{Schema}.
#' @rdname    Schema-class
#' @export
RepertoireSchema <- load_schema("Repertoire")

#' @details   \code{GermlineSetSchema}: AIRR GermlineSet \code{Schema}.
#' @rdname    Schema-class
#' @export
GermlineSetSchema <- load_schema("GermlineSet")

#' @details   \code{GenotypeSetSchema}: AIRR GenotypeSet \code{Schema}.
#' @rdname    Schema-class
#' @export
GenotypeSetSchema <- load_schema("GenotypeSet")

#' @details   \code{AIRRSchema}: named list containing all non-experimental 
#'                               AIRR \code{Schema} objects.
#' @rdname    Schema-class
#' @export
AIRRSchema <- list("Info"=load_schema("InfoObject"),
                   "DataFile"=load_schema("DataFile"),
                   "Alignment"=load_schema("Alignment"),
                   "Rearrangement"=load_schema("Rearrangement"),
                   "Repertoire"=load_schema("Repertoire"),
                   "Ontology"=load_schema("Ontology"),
                   "Study"=load_schema("Study"),
                   "Subject"=load_schema("Subject"),
                   "Diagnosis"=load_schema("Diagnosis"),
                   "SampleProcessing"=load_schema("SampleProcessing"),
                   "CellProcessing"=load_schema("CellProcessing"),
                   "PCRTarget"=load_schema("PCRTarget"),
                   "NucleicAcidProcessing"=load_schema("NucleicAcidProcessing"),
                   "SequencingRun"=load_schema("SequencingRun"),
                   "SequencingData"=load_schema("SequencingData"),
                   "DataProcessing"=load_schema("DataProcessing"),
                   "GermlineSet"=load_schema("GermlineSet"),
                   "Acknowledgement"=load_schema("Acknowledgement"),
                   "RearrangedSequence"=load_schema("RearrangedSequence"),
                   "UnrearrangedSequence"=load_schema("UnrearrangedSequence"),
                   "SequenceDelineationV"=load_schema("SequenceDelineationV"),
                   "AlleleDescription"=load_schema("AlleleDescription"),
                   "GenotypeSet"=load_schema("GenotypeSet"),
                   "Genotype"=load_schema("Genotype"),
                   "Cell"=load_schema("Cell"),
                   "Clone"=load_schema("Clone"))

#### Read TSV ####

#' Read AIRR tabular data
#' 
#' \code{read_tabular} reads a tab-delimited (TSV) file containing tabular AIRR records.
#'
#' @param    file        input file path.
#' @param    base        starting index for positional fields in the input file.
#'                       If \code{"1"}, then these fields will not be modified.
#'                       If \code{"0"}, then fields ending in \code{"_start"} and \code{"_end"}
#'                       are 0-based half-open intervals (python style) in the input file
#'                       and will be converted to 1-based closed-intervals (R style).
#' @param    schema      \code{Schema} object defining the output format.
#' @param    aux_types   named vector or list giving the type for fields that are not
#'                       defined in \code{schema}. The field name is the name, the value
#'                       the type, denoted by one of \code{"c"} (character), \code{"l"} (logical),
#'                       \code{"i"} (integer), \code{"d"} (double), or \code{"n"} (numeric).
#' @param    ...         additional arguments to pass to \link[readr]{read_delim}.
#'
#' @return   A \code{data.frame} of the TSV file with appropriate type and position conversion
#'           for fields defined in the specification.
#'
#' @seealso
#' See \link{Schema} for the AIRR schema object definition.
#' See \link{write_tabular} for writing AIRR data.
#' 
#' @examples
#' # Get path to the rearrangement-example file
#' file <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")
#'
#' # Load data file
#' df <- read_rearrangement(file)
#'
#' @export
read_tabular <- function(file, schema, base=c("1", "0"), aux_types=NULL,...) {
    # Check arguments
    base <- match.arg(base)

    # Define types
    parsers <- c("character"="c", "logical"="l", "integer"="i", "double"="d", "numeric"="n")
    header <- names(suppressMessages(readr::read_tsv(file, n_max=1)))
    schema_fields <- intersect(names(schema), header)
    cast <- setNames(lapply(schema_fields, function(f) parsers[schema[f]$type]), schema_fields)
    cast <- c(cast, list(.default = col_character()))

    if(!is.null(aux_types)){
        aux_types <- aux_types[names(aux_types) %in% header]
        aux_cols <- setNames(lapply(aux_types, function(f) parsers[f]), names(aux_types))
        cast <- c(cast, aux_cols)
    }

    types <- do.call(readr::cols, cast)

    # Read file
    data <- suppressMessages(readr::read_tsv(file, col_types=types, na=c("", "NA", "None"), ...))

    # Validate file
    valid_data <- validate_tabular(data, schema=schema)
    
    # Adjust indexes
    if (base == "0") {
        start_positions <- grep("_start$", names(data), perl=TRUE)
        if (length(start_positions) > 0) {
            data[, start_positions] <- data[, start_positions] + 1
        }
    }

    return(data)
}


#' @details
#' \code{read_rearrangement} reads an AIRR TSV containing Rearrangement data.
#' 
#' @rdname read_tabular
#' @export
read_rearrangement <- function(file, base=c("1", "0"), ...) {
  read_tabular(file, base=base, schema=RearrangementSchema, ...)
}

#' @details
#' \code{read_alignment} reads an AIRR TSV containing Alignment data.
#' 
#' @rdname read_tabular
#' @export
read_alignment <- function(file, base=c("1", "0"), ...) {
  msg <- paste("read_alignment is deprecated and will be removed in a future release.",
               "Use read_tabular with the argument schema=AlignmentSchema instead.",
               "See help(\"Deprecated\")", 
               sep="\n")
  .Deprecated(msg=msg)
  read_tabular(file, base=base, schema=AlignmentSchema, ...)
}


#### Read YAML/JSON ####

#' Read an AIRR Data Model file in YAML or JSON format
#' 
#' \code{read_airr} loads a YAML or JSON file containing AIRR Data Model records.
#'
#' @param    file      path to the input file.
#' @param    format    format of the input file. Must be one of \code{"auto"}, \code{"yaml"}, or 
#'                     \code{"json"}. If \code{"auto"} (default), the format will be 
#'                     detected from the \code{file} extension.
#' @param    validate  run schema validation if \code{TRUE}.
#' @param    model     if \code{TRUE} validate only AIRR DataFile defined objects. If \code{FALSE} 
#'                     attempt validation of all objects in \code{data}.
#'                     Ignored if \code{validate=FALSE}
#' 
#' @return   A named nested \code{list} contained in the AIRR Data Model with the top-level
#'           names reflecting the individual AIRR objects.
#'                   
#' @seealso  
#' See \link{Schema} for the AIRR schema definition objects.
#' See \link{write_airr} for writing AIRR Data Model records in YAML or JSON format.
#' 
#' @examples
#' # Get path to the Reportoire and GermlineSet example files
#' f1 <- system.file("extdata", "repertoire-example.yaml", package="airr")
#' f2 <- system.file("extdata", "germline-example.json", package="airr")
#' 
#' # Load data files
#' repertoire <- read_airr(f1)
#' germline <- read_airr(f2)
#' 
#' @export
read_airr <- function(file, format=c("auto", "yaml", "json"), validate=TRUE, model=TRUE) {
    # Check arguments
    format <- match.arg(format)
    
    # Autodetect format
    if (format == "auto") { format <- tolower(tools::file_ext(file)) }
    
    # Load data
    if (format == "yaml") {
        records <- read_airr_yaml(file, validate=validate, model=model)
    } else if (format == "json") {
        records <- read_airr_json(file, validate=validate, model=model)
    } else {
        stop("Unrecognized file extension ", format, "; must be either .json or .yaml.")
    }
    
    # Return
    return(records)
}

# Read an AIRR YAML file
# 
# \code{read_airr_yaml} loads a YAML file containing AIRR Data Model records.
#
# @param    file      path to the YAML input file.
# @param    validate  run schema validation if \code{TRUE}.
# @param    model     if \code{TRUE} validate only AIRR DataFile defined objects. If \code{FALSE} 
#                     attempt validation of all objects in \code{data}.
#                     Ignored if \code{validate=FALSE}
#                      
# @return   A named nested \code{list} contained in the AIRR Data Model with the top-level
#           names reflecting the individual AIRR objects.
#                   
# @seealso  
# See \link{Schema} for the AIRR schema definition objects.
# See \link{write_airr_yaml} for writing AIRR data in YAML format.
# 
# @examples
# # Get path to the repertoire-example file
# file <- system.file("extdata", "repertoire-example.yaml", package="airr")
# 
# # Load data file
# repr <- read_airr_yaml(file)
read_airr_yaml <- function(file, validate=TRUE, model=TRUE) {
  
  # YAML format
  data <- yaml::read_yaml(file)

  # Validation. Warnings are thrown for fields for AIRR compliance failures
  if (validate) {
      valid <- validate_airr(data, model=model)
  }
  
  return(data)
}


# Read an AIRR JSON file
# 
# \code{read_airr_json} loads a JSON file containing AIRR Data Model records.
#
# @param    file      path to the JSON input file.
# @param    validate  run schema validation if \code{TRUE}.
# @param    model     if \code{TRUE} validate only AIRR DataFile defined objects. If \code{FALSE} 
#                     attempt validation of all objects in \code{data}.
#                     Ignored if \code{validate=FALSE}
#
# @return   A named nested \code{list} contained in the AIRR Data Model with the top-level
#           names reflecting the individual AIRR objects.
#                   
# @seealso  
# See \link{Schema} for the AIRR schema object definition.
# See \link{write_airr_json} for writing AIRR data in JSON format.
# 
# @examples
# # Get path to the rearrangement-example file
# file <- system.file("extdata", "germline-example.json", package="airr")
# 
# # Load data file
# repr <- read_airr_json(file)
read_airr_json <- function(file, validate=TRUE, model=TRUE) {
  
  # Read JSON format
  data <- jsonlite::fromJSON(file, 
                             simplifyVector=TRUE,
                             simplifyMatrix=FALSE,
                             simplifyDataFrame=FALSE,
                             flatten=FALSE)
  
  # Validation. Warnings are thrown for fields for AIRR compliance failures
  if (validate) {
      valid <- validate_airr(data, model=model)
  }
  
  return(data)
}

#### Write TSV ####

#' Write an AIRR tabular data
#' 
#' \code{write_tabular} writes a TSV containing AIRR tabular records.
#'
#' @param    data    \code{data.frame} of Rearrangement data.
#' @param    file    output file name.
#' @param    base    starting index for positional fields in the output file.
#'                   Fields in the input \code{data} are assumed to be 1-based
#'                   closed-intervals (R style).
#'                   If \code{"1"}, then these fields will not be modified.
#'                   If \code{"0"}, then fields ending in \code{_start} and \code{_end}
#'                   will be converted to 0-based half-open intervals (python style)
#'                   in the output file.
#' @param    schema  \code{Schema} object defining the output format.
#' @param    ...     additional arguments to pass to \link[readr]{write_delim}.
#'
#' @return   NULL
#'
#' @seealso
#' See \link{Schema} for the AIRR schema object definition.
#' See \link{read_tabular} for reading to AIRR files.
#' 
#' @examples
#' # Get path to the rearrangement-example file
#' file <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")
#'
#' # Load data file
#' df <- read_rearrangement(file)
#'
#' # Write a Rearrangement data file
#' outfile <- file.path(tempdir(), "output.tsv")
#' write_tabular(df, outfile, schema=RearrangementSchema)
#'
#' @export
write_tabular <- function(data, file, schema, base=c("1", "0"),  ...) {
    ## DEBUG
    # data <- data.frame("sequence_id"=1:4, "extra"=1:4, "a"=LETTERS[1:4])

    data_name <- deparse(substitute(data))
    schema_name <- deparse(substitute(schema))

    # Check arguments
    base <- match.arg(base)

    # Fill in missing required columns
    missing <- setdiff(schema@required, names(data))
    if (length(missing) > 0 ) {
        data[, missing] <- NA
    }

    # order columns
    ordering <- c(intersect(names(schema), names(data)),
                  setdiff(names(data), names(schema)))
    data <- data[, ordering]

    # Adjust indexes
    if (base == "0") {
        start_positions <- grep("_start$", names(data), perl=TRUE)
        if (length(start_positions) > 0) {
            data[, start_positions] <- data[, start_positions] - 1
        }
    }
    valid <- suppressWarnings(validate_tabular(data, schema))
    if (!valid) {
        w <- names(warnings())
        w <- gsub("Warning: *", "" ,w)
        err_msg <- paste0(data_name, " is not a valid ", schema_name, "\n")
        err_msg <- paste(err_msg, paste(w, collapse = "\n"))
        warning(err_msg)
    }


    # write logical fields as T/F
    logical_fields <- names(which(sapply(schema@properties,
                                         '[[', "type") == "logical"))
    logical_fields <- intersect(colnames(data), logical_fields)
    if (length(logical_fields) > 0 ) {
        for (log_field in logical_fields) {
            logical_values <- data[[log_field]] %in% c(TRUE, FALSE)
            data[[log_field]] <- as.character(data[[log_field]])
            if (length(logical_values) > 0 ) {
                data[[log_field]][logical_values] <- c("T", "F")[match(data[[log_field]][logical_values],
                                                            c("TRUE", "FALSE"))]
            }
        }
    }

    # Write
    write_tsv(data, file, na="", ...)
}


#' @details
#' \code{write_rearrangement} writes a \code{data.frame} containing AIRR Rearrangement data to TSV.
#' 
#' @rdname write_tabular
#' @export
write_rearrangement <- function(data, file, base=c("1", "0"), ...) {
    write_tabular(data, file, base=base, schema=RearrangementSchema, ...)
}


#' @details
#' \code{write_alignment} writes a \code{data.frame} containing AIRR Alignment data to TSV.
#' 
#' @rdname write_tabular
#' @export
write_alignment <- function(data, file, base=c("1", "0"), ...) {
    msg <- paste("write_alignment is deprecated and will be removed in a future release.",
                 "Use write_tabular with the argument schema=AlignmentSchema instead.",
                 "See help(\"Deprecated\")", 
                 sep="\n")
    .Deprecated(msg=msg)
    write_tabular(data, file, base=base, schema=AlignmentSchema, ...)
}

#### Write YAML/JSON ####

#' Write AIRR Data Model records to YAML or JSON files
#' 
#' \code{write_airr} writes a YAML or JSON file containing AIRR Data Model records.
#'
#' @param    data      \code{list} containing AIRR Model Records.
#' @param    file      output file name.
#' @param    format    format of the output file. Must be one of \code{"auto"}, \code{"yaml"}, or 
#'                     \code{"json"}. If \code{"auto"} (default), the format will be 
#'                     detected from the \code{file} extension.
#' @param    validate  run schema validation prior to write if \code{TRUE}.
#' @param    model     if \code{TRUE} validate and write only AIRR DataFile defined objects. 
#'                     If \code{FALSE} attempt validation and write of all objects in \code{data}.
#' 
#' @seealso
#' See \link{Schema} for the AIRR schema definition objects.
#' See \link{read_airr} for reading to AIRR Data Model files.
#' 
#' @examples
#' # Get path to the repertoire-example file
#' file <- system.file("extdata", "repertoire-example.yaml", package="airr")
#' 
#' # Load data file
#' repertoire <- read_airr(file)
#' 
#' # Write a Rearrangement data file
#' outfile <- file.path(tempdir(), "output.yaml")
#' write_airr(repertoire, outfile)
#' 
#' @export
write_airr <- function(data, file, format=c("auto", "yaml", "json"), validate=TRUE, model=TRUE) {
    # Check arguments
    format <- match.arg(format)
    
    # Autodetect format
    if (format == "auto") { format <- tolower(tools::file_ext(file)) }
    
    # Write data
    if (format == "yaml") {
        write_airr_yaml(data, file, validate=validate, model=model)
    } else if (format == "json") {
        write_airr_json(data, file, validate=validate, model=model)
    } else {
        stop("Unrecognized file extension ", format, "; must be either .json or .yaml.")
    }
}


# Write an AIRR yaml
# 
# \code{write_airr_yaml} writes a yaml containing AIRR formatted records.
#
# @param    data      object containing Repertoire data.
# @param    file      output file name.
# @param    validate  run schema validation prior to write if \code{TRUE}.
# @param    model     if \code{TRUE} validate only AIRR DataFile defined objects. If \code{FALSE} 
#                     attempt validation of all objects in \code{data}.
# 
# @seealso
# See \link{Schema} for the AIRR schema object definition.
# See \link{read_airr_yaml} for reading to AIRR files.
# 
# @examples
# # Get path to the rearrangement-example file
# file <- system.file("extdata", "repertoire-example.yaml", package="airr")
# 
# # Load data file
# repr <- read_airr(file)
# 
# # Write a Rearrangement data file
# outfile <- file.path(tempdir(), "output.yaml")
# write_airr_yaml(repr, outfile)
write_airr_yaml <- function(data, file, validate=TRUE, model=TRUE) {
    # Validate prior to write
    if (validate) {
        valid <- validate_airr(data, model=model)
    }
    
    # Subset to AIRR DataFile records
    if (model) {
        data <- data[names(data) %in% names(DataFileSchema@properties)]
    }
    
    # Write
    yaml::write_yaml(data, file)
}

# Write an AIRR json
# 
# \code{write_airr_json} writes a yaml containing AIRR formatted records.
#
# @param    data    object containing Repertoire data.
# @param    file    output file name.
# @param    validate  run schema validation prior to write if \code{TRUE}.
#
# @return   NULL
# 
# @seealso
# See \link{Schema} for the AIRR schema object definition.
# See \link{read_airr_json} for reading to AIRR files.
# 
# @examples
# # Get path to the rearrangement-example file
# file <- system.file("extdata", "germline-example.json", package="airr")
# 
# # Load data file
# repr <- read_airr(germline)
# 
# # Write a Rearrangement data file
# outfile <- file.path(tempdir(), "output.json")
# write_airr_json(repr, outfile)
write_airr_json <- function(data, file, validate=TRUE, model=TRUE) {
    # Validate prior to write
    if (validate) {
        valid <- validate_airr(data, model=model)
    }
  
    # Subset to AIRR DataFile records
    if (model) {
        data <- data[names(data) %in% names(DataFileSchema@properties)]
    }
    
    # Write
    json <- jsonlite::toJSON(data, auto_unbox=TRUE, null="null", na="null")
    write(json, file)
}


#### Validation ####

#' Validate tabular AIRR data
#' 
#' \code{validate_tabular} validates compliance of the contents of a \code{data.frame} 
#' to the AIRR standards.
#'
#' @param    data    \code{data.frame} of tabular data to validate.
#' @param    schema  \code{Schema} object defining the data standard of the table.
#'
#' @return   Returns \code{TRUE} if the input \code{data} is compliant and
#'           \code{FALSE} if not.
#' 
#' @examples
#' # Get path to the rearrangement-example file
#' file <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")
#'
#' # Load data file
#' df <- read_rearrangement(file)
#'
#' # Validate a data.frame against the Rearrangement schema
#' validate_rearrangement(df)
#' 
#' @export
validate_tabular <- function(data, schema) {
    # Initialize return value
    valid <- TRUE
    
    # Check all required fields exist
    missing_fields <- setdiff(schema@required, names(data))
    
    if (length(missing_fields) > 0 ) {
        valid <- FALSE
        warning(paste("Warning: File is missing AIRR mandatory field(s):",
                      paste(missing_fields, collapse = ", ")))
    }
    
    # Validate sequence_id: 
    # - uniqueness
    # - not empty
    if ("sequence_id" %in% colnames(data)) {
        dup_ids <- duplicated(data[['sequence_id']])
        if (any(dup_ids)) {
            valid <- FALSE
            warning(paste("Warning: sequence_id(s) are not unique:",
                          paste(data[['sequence_id']][dup_ids], collapse = ", ")))
        }
        empty_rows <- which(data[['sequence_id']] %in% c("None", "", NA))
        if (length(empty_rows) > 0 ) {
            # TODO
            # valid <- FALSE
            warning(paste("Warning: sequence_id is empty for row(s):",
                          paste(empty_rows, collapse = ", ")))           
        }
    }
    
    # check logical fields
    logical_fields <- names(which(sapply(schema@properties, 
                                         '[[', "type") == "logical"))
    logical_fields <- intersect(colnames(data), logical_fields)
    if (length(logical_fields) > 0 ) {
        for (log_field in logical_fields) {
            not_logical <- data[[log_field]] %in% c(TRUE, FALSE) == FALSE
            if (any(not_logical)) {
                warning(paste("Warning:",log_field,"is not logical for row(s):",
                              paste(which(not_logical), collapse = ", ")))            
            } else {
                NULL
            }
        }
    }
    
    return(valid)
}

#' @details
#' \code{validate_rearrangement} validates the standards compliance of AIRR Rearrangement 
#' data stored in a \code{data.frame}
#' 
#' @rdname validate_tabular
#' @export
validate_rearrangement <- function(data) {
    validate_tabular(data, schema=RearrangementSchema)
}


#' Validate an AIRR Data Model nested list representation
#' 
#' \code{validate_airr} validates the fields in a named nested list representation of the 
#' AIRR Data Model. Typically, generating by reading of JSON or YAML formatted AIRR files.
#'
#' @param    data     \code{list} containing records of an AIRR Data Model objected imported from 
#'                    a YAML or JSON representation.
#' @param    model    if \code{TRUE} validate only AIRR DataFile defined objects. If \code{FALSE} 
#'                    attempt validation of all objects in \code{data}.
#' @param    each     if \code{TRUE} return a logical vector with results for each object in \code{data}
#'                    instead of a single \code{TRUE} or \code{FALSE} value.
#' 
#' @return   Returns \code{TRUE} if the input \code{data} is compliant with AIRR standards and
#'           \code{FALSE} if not. If \code{each=TRUE} is set, then a vector with results for each
#'           each object in \code{data} is returned instead.
#'                   
#' @seealso  
#' See \link{Schema} for the AIRR schema definitions.
#' See \link{read_airr} for loading AIRR Data Models from a file.
#' See \link{write_airr} for writing AIRR Data Models to a file.
#' 
#' @examples
#' # Get path to the rearrangement-example file
#' f1 <- system.file("extdata", "repertoire-example.yaml", package="airr")
#' f2 <- system.file("extdata", "germline-example.json", package="airr")
#' 
#' # Load data file
#' repertoire <- read_airr(f1)
#' germline <- read_airr(f2)
#' 
#' # Validate a single record
#' validate_airr(repertoire)
#' 
#' # Return validation for individual objects
#' validate_airr(germline, each=TRUE)
#' 
#' @export
validate_airr <- function(data, model=TRUE, each=FALSE) {
    # This is a wrapper function to allow recursive validation of the different entries in yaml file
    # Directly calling validate_entry does not work, because the function
    # validate_entry also needs to work for recursive calling of reference schemes
    
    # Iterate through objects in input data
    valid_sum <- logical()
    for (n in names(data)) {
        if (n %in% c('Info', 'DataFile')) { next }
        entry <- data[[n]]
        if (is.null(entry)) { next }

        # Check for non-DataFile objects
        if (model && !(n %in% names(DataFileSchema@properties))) { 
            warning('Skipping validation of non-DataFile object: ', n)
            next 
        }
        
        # Load schema
        if (n %in% names(AIRRSchema)) { 
            schema <- AIRRSchema[[n]] 
        } else { 
            schema <- tryCatch(load_schema(n), error=function(e) NULL)
        }
        
        # Fail invalid schema
        if (is.null(schema)) {
            warning('Unrecognized schema: ', n)
            valid <- FALSE
        } else {
            # Recursively validate all entries
            valid <- sapply(entry, validate_entry, schema=schema)
        }
        
        # Store check result
        valid_sum <- append(setNames(all(valid), n), valid_sum)
    }
    
    # Data only valid if all entries valid
    if (!each) { valid_sum <- all(valid_sum) }
    
    return(valid_sum)
}


# Validation function for a single entry in the yaml file
#
# @param    entry    AIRR data in a nested list structure
# @param    schema             Schema definition object
# @returns  TRUE or FALSE
validate_entry <- function(entry, schema) {
    schema_name <- schema@definition
    valid <- TRUE
    
    # Check all required fields exist
    missing_fields <- setdiff(schema@required, names(entry))
    
    if (length(missing_fields) > 0 ) {
        valid <- FALSE
        warning(paste("Warning:", schema_name, "object is missing AIRR mandatory field(s):",
                      paste(missing_fields, collapse = ", "), "\n"))
    }
    
    # loop through all fields in the list and check if they refer to other schemes
    for(f in names(entry)) {
        # get the reference scheme
        reference_schemes <- schema[f]$ref
        
        # simple recursive (reference scheme in 1st level)
        # in this case the type on the 1st level is NULL
        if (is.na(schema[f][["type"]]) || is.null(schema[f][["type"]])) {
            if (!is.null(reference_schemes)) {
                v <- validate_entry(entry[[f]], schema=reference_schemes)
                if (!v) { valid <- FALSE }
            }
        # entry of array type with a list of on or several reference schemes
        } else if (schema[f][["type"]] == "array" & !is.null(reference_schemes)) {
            # this is the number of different reference schemes
            n_schemes <- length(reference_schemes)
            # this is the number of elements in the array
            n_array_entries <- length(entry[[f]])
            # loop over all reference schemes in list
            for (n_ref in seq_len(n_schemes)) {
                # recursively validate the entries in the array
                for (n_array in seq_len(n_array_entries)) {
                    v <- validate_entry(entry[[f]][[n_array]], schema = reference_schemes[[n_ref]])
                    if (!v) { valid <- FALSE }
                }
            }
        # check if the entry type is correct
        } else if (class(entry[[f]]) != schema[f][["type"]]) {
            # one reason for non-identical types can be that the entry is nullable
            nullable <- schema[f][["x-airr"]][["nullable"]]
            # if not specified, it should be nullable
            if (is.null(nullable)) { nullable <- TRUE }
            if (!(nullable & is.null(entry[[f]]))) {
                # another reason for types not matching is the array format
                # we test whether the entries are lists
                if (!(schema[f][["type"]] == "array" & is.vector(entry[[f]]))) {
                    # another reason for types not matching is the numeric arguments being read as integers
                    # we test whether the entries are numeric
                    if (!(schema[f][["type"]] == "numeric" & is.numeric(entry[[f]]))) {
                        valid <- FALSE
                        warning(paste("Warning:", schema_name, "entry does not have the required type",
                                      schema[f][["type"]], ":", f, "\n"))
                    }  
                }
            }
        }
    }
    
    # return to indicate whether entries are valid
    return(valid)
}

#### Setup ####

# Installed package context
parent_path <- ".."

# Local script context
# library(testthat)
# library(airr)
# parent_path <- "tests"

# Rearrangement test files
good_rearrangement_file <- file.path(parent_path, "data-tests", "good_rearrangement.tsv")
bad_rearrangement_file <- file.path(parent_path, "data-tests", "bad_rearrangement.tsv")

# Repertoire test files
good_repertoire_file <- file.path(parent_path, "data-tests", "good_repertoire.yaml")
bad_repertoire_file <- file.path(parent_path, "data-tests", "bad_repertoire.yaml")
warn_repertoire_file <- file.path(parent_path, "data-tests", "warn_repertoire.json")

# Germline test files
good_germline_set_file <- file.path(parent_path, "data-tests", "good_germline_set.json")
bad_germline_set_file <- file.path(parent_path, "data-tests", "bad_germline_set.json")
good_genotype_set_file <- file.path(parent_path, "data-tests", "good_genotype_set.json")
bad_genotype_set_file <- file.path(parent_path, "data-tests", "bad_genotype_set.json")

# Combined test files
good_combined_yaml <- file.path(parent_path, "data-tests", "good_combined_airr.yaml")
good_combined_json <- file.path(parent_path, "data-tests", "good_combined_airr.json")
good_combined_names <- c("Repertoire", "GermlineSet", "GenotypeSet")

# Expected warnings for bad_rearrangement_file
expected_w <- c(
    "Warning: File is missing AIRR mandatory field(s): sequence",
    "Warning: sequence_id(s) are not unique: IVKNQEJ01AJ44V, IVKNQEJ01AJ44V",
    "Warning: sequence_id is empty for row(s): 7",
    "Warning: rev_comp is not logical for row(s): 4",
    "Warning: productive is not logical for row(s): 1"
)


#### Rearrangement ####

context("Rearrangement I/O - good data")

test_that("read_rearrangement loads a data.frame", {
    tbl_1 <- read_rearrangement(good_rearrangement_file)
    expect_true(is.data.frame(tbl_1))
})

test_that("read_tabular loads a data.frame", {
    expect_error(read_tabular(good_rearrangement_file))
    tbl_1 <- read_tabular(good_rearrangement_file, schema=RearrangementSchema)
    expect_true(is.data.frame(tbl_1))
})

test_that("read_tabular applies base", {
    tbl_0 <- read_tabular(good_rearrangement_file, base="0", schema=RearrangementSchema)
    tbl_1 <- read_tabular(good_rearrangement_file, base="1", schema=RearrangementSchema)
    expect_true(is.data.frame(tbl_0))
    expect_true(validate_tabular(tbl_0, schema = RearrangementSchema))
    start_positions <- grep("_start$", names(tbl_0), perl=TRUE)
    expect_equivalent(tbl_0[, start_positions] - 1, tbl_1[, start_positions])
})

test_that("read_tabular reads zipped tsv", {
  # Create test data gzip
  tmp_file <- tempfile(fileext = ".tsv.gz")

  x <- readr::read_tsv(good_rearrangement_file)
  readr::write_tsv(x, file=tmp_file)

  tbl_1 <- read_tabular(tmp_file, schema=RearrangementSchema)
  expect_true(is.data.frame(tbl_1))

  # bzip2
  tmp_file <- tempfile(fileext=".tsv.bz2")
  readr::write_tsv(x, file=tmp_file)

  tbl_1 <- read_tabular(tmp_file, schema=RearrangementSchema)
  expect_true(is.data.frame(tbl_1))
})

test_that("Columns are of expected type", {
    # Create test data
    tmp_file <- tempfile(fileext = ".tsv")

    tbl_0 <- read_tabular(good_rearrangement_file, schema=RearrangementSchema)

    # Create non-schema  columns
    extra_cols <- data.frame(
        extra.int = c(1:nrow(tbl_0)),
        extra.numeric = rnorm(nrow(tbl_0)),
        extra.double = rnorm(nrow(tbl_0)),
        extra.character = stringi::stri_rand_strings(nrow(tbl_0), 6)
    )
    extra_cols_na <- data.frame(
        extra.int = NA,
        extra.numeric = NA,
        extra.double = NA,
        extra.character = NA
    )

    # Get the number of replications to create a data.frame with NAs to make
    # guessing of column type fail (all NAs)
    num_rep <- ceiling(1000/nrow(tbl_0)) + 1
    tbl_1 <- do.call(
        "rbind",
        replicate(num_rep,
                  cbind(tbl_0, extra_cols_na),
                  simplify = FALSE))
    # Make all sequence_id unique
    tbl_1$sequence_id <- c(1:nrow(tbl_1))

    tbl_0 <- cbind(tbl_0, extra_cols)
    tbl_0 <- rbind(tbl_1, tbl_0)
    write_tabular(tbl_0, tmp_file, schema=RearrangementSchema)

    # Read with col_character() as default
    expect_is(tbl_0 <- read_tabular(tmp_file, schema=RearrangementSchema), "data.frame")
    expect_is(tbl_0$extra.int, "character")
    expect_is(tbl_0$extra.numeric, "character")
    expect_is(tbl_0$extra.character, "character")

    expect_is(read_tabular(
        good_rearrangement_file, schema=RearrangementSchema,
        aux_types =
            c(extra.int = "integer",
              extra.double = "double",
              extra.numeric = "numeric",
              extra.character = "character")), "data.frame")

    expect_is(tbl_0 <- read_tabular(
        tmp_file, "1", schema = RearrangementSchema,
        aux_types =
            c(extra.int = "integer",
              extra.double = "double",
              extra.numeric = "numeric",
              extra.character = "character")), "data.frame")
    expect_is(tbl_0$extra.int, "integer")
    expect_is(tbl_0$extra.double, "numeric")
    expect_is(tbl_0$extra.numeric, "numeric")
    expect_is(tbl_0$extra.character, "character")
})


test_that("write_tabular writes a file with logicals encoded T/F", {
    tbl <- read_tabular(good_rearrangement_file, schema=RearrangementSchema)
    out_file <- file.path(tempdir(), "test_out.tsv")
    write_tabular(tbl, out_file, schema = RearrangementSchema)
    expect_true(file.exists(out_file))
    reload_tbl <- read.delim(out_file, colClasses="character")
    expect_true(all(reload_tbl[['rev_comp']] == "T"))
    expect_equal(reload_tbl[['productive']],
                 c("T","T","F","T","T","F","F","F","T"))
})

context("Rearrangement I/O - bad data")

test_that("read_tabular with bad data", {
    # Expect valid==FALSE
    bad_data <- suppressWarnings(read_tabular(bad_rearrangement_file, base="1", schema=RearrangementSchema))
    expect_false(suppressWarnings(validate_tabular(bad_data, schema=RearrangementSchema)))
    # Check error messages
    w <- capture_warnings(validate_tabular(bad_data, schema=RearrangementSchema))
    expect_equal(w, expected_w)
})

test_that("write_tabular writes a bad file, with warnings, with logicals T/T", {
    bad_data <- suppressWarnings(read_tabular(bad_rearrangement_file, base="1", schema=RearrangementSchema))
    out_file <- file.path(tempdir(), "test_out.tsv")
    expect_warning(write_tabular(bad_data, out_file, schema = RearrangementSchema))
    expect_true(file.exists(out_file))
    reload_tbl <- read.delim(out_file, colClasses="character")
    expect_equal(reload_tbl[['rev_comp']],
                c("T","T","T","","T","T","T","T","T","T","T"))
    expect_equal(reload_tbl[['productive']],
                 c("","T","F","T","T","F","F","F","T","T","T"))
})

#### Repertoire ####

context("Repertoire I/O - good data")

test_that("read_airr loads a Repertoire", {
    rep_1 <- read_airr(good_repertoire_file)
    expect_true(is.list(rep_1))
})

test_that("read_airr with format=yaml loads a Repertoire", {
    rep_1 <- read_airr(good_repertoire_file, format="yaml")
    expect_true(is.list(rep_1))
})

test_that("write_airr writes a yaml Repertoire", {
    geno <- read_airr(good_repertoire_file, validate=T)
    
    tmp_file <- tempfile(fileext=".yaml")
    write_airr(geno, tmp_file, validate=T, model=F)
    geno_tmp <- read_airr(tmp_file, validate=T, model=F)
    expect_identical(geno, geno_tmp)
})

test_that("write_airr writes a json Repertoire", {
    geno <- read_airr(good_repertoire_file, validate=T)
    
    tmp_file <- tempfile(fileext=".json")
    write_airr(geno, tmp_file, validate=T, model=F)
    geno_tmp <- read_airr(tmp_file, validate=T, model=F)
    expect_identical(geno, geno_tmp)
})

context("Repertoire I/O - bad data")

test_that("validate_airr with bad data returns an error", {
    bad_data <- expect_warning(read_airr(bad_repertoire_file))
    expect_false(suppressWarnings(validate_airr(bad_data)))
})

#### GermlineSet ####

context("GermlineSet I/O - good data")

# TODO: Update good data so validation passes
test_that("read_airr loads a GermlineSet", {
  rep_1 <- read_airr(good_germline_set_file, validate=T)
  expect_true(is.list(rep_1))
})

context("GermlineSet I/O - bad data")

test_that("validate_airr with bad data returns an error", {
  bad_data <- read_airr(bad_germline_set_file, validate=F)
  expect_false(expect_warning(validate_airr(bad_data)))
})

#### GenotypeSet ####

context("GenotypeSet I/O - good data")

# TODO: Update good data so validation passes
test_that("read_airr loads a GenotypeSet", {
  rep_1 <- read_airr(good_genotype_set_file, validate=T)
  expect_true(is.list(rep_1))
})

context("GenotypeSet I/O - bad data")

test_that("validate_airr with bad data returns an error", {
  bad_data <- read_airr(bad_genotype_set_file, validate=F)
  expect_false(expect_warning(validate_airr(bad_data)))
})

#### Combined Data ####

context("Combined I/O")

test_that("read_airr loads a combined yaml file", {
    repr <- read_airr(good_combined_yaml, validate=T)
    expect_true(is.list(repr))
    expect_equal(names(repr), good_combined_names)
})

test_that("read_airr loads a combined json file", {
    repr <- read_airr(good_combined_json, validate=T)
    expect_true(is.list(repr))
    expect_equal(names(repr), good_combined_names)
})

test_that("read_airr json and yaml load identical objects", {
    repr_yaml <- read_airr(good_combined_yaml, validate=T)
    repr_json <- read_airr(good_combined_json, validate=T)
    expect_identical(repr_yaml, repr_json)
})

test_that("write_airr writes a combined yaml file", {
    repr <- read_airr(good_combined_yaml, validate=T)
    
    tmp_file <- tempfile(fileext=".yaml")
    write_airr(repr, tmp_file, validate=T, model=F)
    repr_tmp <- read_airr(tmp_file, validate=T, model=F)
    expect_identical(repr, repr_tmp)
})

test_that("write_airr writes a combined json file", {
    repr <- read_airr(good_combined_json, validate=T)
    
    tmp_file <- tempfile(fileext=".json")
    write_airr(repr, tmp_file, validate=T, model=F)
    repr_tmp <- read_airr(tmp_file, validate=T, model=F)
    expect_identical(repr, repr_tmp)
})

test_that("write_airr json and yaml writes identical objects", {
    repr <- read_airr(good_combined_yaml, validate=T)
    
    tmp_yaml <- tempfile(fileext=".yaml")
    write_airr(repr, tmp_yaml, validate=T)
    repr_yaml <- read_airr(tmp_yaml, validate=T)
    
    tmp_json <- tempfile(fileext=".json")
    write_airr(repr, tmp_json, validate=T)
    repr_json <- read_airr(tmp_json, validate=T)
    
    expect_identical(repr_yaml, repr_json)
})

#### DataFile ####

context("DataFile")

test_that("read_airr works with DataFile", {
    rep1 <- read_airr(good_combined_yaml, validate=T, model=F)
    rep2 <- read_airr(good_combined_yaml, validate=T, model=T)
    expect_identical(rep1, rep2)
})

test_that("validate_airr works with DataFile", {
    rep <- read_airr(good_combined_yaml, validate=F, model=F)
    rep[["Nonsense"]] <- list(list(id="id", data="data"))
    expect_true(expect_warning(validate_airr(rep, model=T),
                               "Skipping validation of non-DataFile object: Nonsense"))
    expect_false(expect_warning(validate_airr(rep, model=F),
                                "Unrecognized schema: Nonsense"))
})

test_that("write_airr yaml works with DataFile", {
    # Test data
    repr <- read_airr(good_combined_yaml, validate=F, model=F)
    repr_extra <- repr
    repr_extra[["Nonsense"]] <- list(list(id="id", data="data"))

    # Write without DataFile constraint
    tmp_file <- tempfile(fileext=".yaml")
    expect_warning(write_airr(repr_extra, tmp_file, validate=T, model=F),
                   "Unrecognized schema: Nonsense")
    repr_tmp <- expect_warning(read_airr(tmp_file, validate=T, model=T),
                               "Skipping validation of non-DataFile object: Nonsense")
    expect_true("Nonsense" %in% names(repr_tmp))
    expect_identical(repr_extra, repr_tmp)
    
    # Write with DataFile constraint
    expect_warning(write_airr(repr_extra, tmp_file, validate=T, model=T),
                   "Skipping validation of non-DataFile object: Nonsense")
    repr_tmp <- read_airr(tmp_file, validate=T, model=T)
    expect_false("Nonsense" %in% names(repr_tmp))
    expect_identical(repr, repr_tmp)
})

test_that("write_airr json works with DataFile", {
    # Test data
    repr <- read_airr(good_combined_json, validate=F, model=F)
    repr_extra <- repr
    repr_extra[["Nonsense"]] <- list(list(id="id", data="data"))
    
    # Write without DataFile constraint
    tmp_file <- tempfile(fileext=".json")
    expect_warning(write_airr(repr_extra, tmp_file, validate=T, model=F),
                   "Unrecognized schema: Nonsense")
    repr_tmp <- expect_warning(read_airr(tmp_file, validate=T, model=T),
                               "Skipping validation of non-DataFile object: Nonsense")
    expect_true("Nonsense" %in% names(repr_tmp))
    expect_identical(repr_extra, repr_tmp)
    
    # Write with DataFile constraint
    expect_warning(write_airr(repr_extra, tmp_file, validate=T, model=T),
                   "Skipping validation of non-DataFile object: Nonsense")
    repr_tmp <- read_airr(tmp_file, validate=T, model=T)
    expect_false("Nonsense" %in% names(repr_tmp))
    expect_identical(repr, repr_tmp)
})

#' @import      methods
#' @importFrom  readr     read_tsv write_tsv 
#'                        cols col_character col_logical col_integer col_double
#' @importFrom  stats     setNames
#' @importFrom  stringi   stri_trim
#' @importFrom  jsonlite  fromJSON toJSON
#' @importFrom  tools     file_ext
#' @importFrom  yaml      read_yaml write_yaml
NULL

---
title: 'AIRR Data Representation Reference Library Usage'
author: "The AIRR Community"
date: '`r Sys.Date()`'
output:
  html_document:
    standalone: true
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    theme: readable
    toc: yes 
  md_document:
    fig_height: 4
    fig_width: 7.5
    preserve_yaml: no
    toc: yes
  pdf_document:
    dev: pdf
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    toc: yes
geometry: margin=1in
fontsize: 11pt
vignette: >
  %\VignetteIndexEntry{Usage Vignette}
  %\usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
---

## Introduction

Since the use of High-throughput sequencing (HTS) was first introduced to analyze 
immunoglobulin (B-cell receptor, antibody) and T-cell receptor repertoires 
(Freeman et al, 2009; Robins et al, 2009; Weinstein et al, 2009), 
the increasing number of studies making use of this technique has produced enormous 
amounts of data and there exists a pressing need to develop and adopt common standards, 
protocols, and policies for generating and sharing data sets. The [Adaptive Immune 
Receptor Repertoire (AIRR) Community](https://airr-community.org) formed in 
2015 to address this challenge (Breden et al, 2017) and has stablished the set of 
minimal metadata elements (MiAIRR) required for describing published AIRR 
datasets (Rubelt et al, 2017) as well as file formats to represent this data in a 
machine-readable form. The `airr` R package provide read, write and validation of data 
following the AIRR Data Representation schemas. This vignette provides a set of 
simple use examples.

### AIRR Data Standards

The AIRR Community's recommendations for a minimal set of metadata that should be 
used to describe an AIRR-seq data set when published or deposited in a 
AIRR-compliant public repository are described in Rubelt et al, 2017. The primary 
aim of this effort is to make published AIRR datasets FAIR (findable, accessible, 
interoperable, reusable); with sufficient detail such that a person skilled in the 
art of AIRR sequencing and data analysis will be able to reproduce the 
experiment and data analyses that were performed.

Following this principles, V(D)J reference alignment annotations are 
saved in standard tab-delimited files (TSV) with associated metadata 
provided in accompanying YAML formatted files. The column names 
and field names in these files have been defined by the 
AIRR Data Representation Working Group using a controlled vocabulary of 
standardized terms and types to refer to each piece of information.

## Reading AIRR formatted files

The `airr` package contains the function `read_rearrangement` to read and validate
files containing AIRR Rearrangement records, where a Rearrangement record describes 
the collection of optimal annotations on a single sequence that has undergone V(D)J reference alignment. 
The usage is straightforward, as the file format is a typical tabulated file. 
The argument that needs attention is `base`, with possible values `"0"` and `"1"`. 
`base` denotes the starting index for positional fields in the input file. 
Positional fields are those that contain alignment coordinates and names 
ending in "_start" and "_end". If the input file is using 1-based closed 
intervals (R style), as defined by the standard, then positional fields will 
not be modified under the default setting of `base="1"`. If the input file is using 
0-based coordinates with half-open intervals (python style), then 
positional fields may be converted to 1-based closed 
intervals using the argument `base="0"`.

### Reading Rearrangements
```{r}
# Imports
library(airr)
library(tibble)

# Read Rearrangement example file
f1 <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")
rearrangement <- read_rearrangement(f1)
glimpse(rearrangement)
```

### Reading AIRR Data Models

AIRR Data Model records, such as Repertoire and GermlineSet, can be read from either a YAML 
or JSON formatted file into a nested list.

```{r}
# Read Repertoire example file
f2 <- system.file("extdata", "repertoire-example.yaml", package="airr")
repertoire <- read_airr(f2)
glimpse(repertoire)

# Read GermlineSet example file
f3 <- system.file("extdata", "germline-example.json", package="airr")
germline <- read_airr(f3)
glimpse(germline)
```


## Writing AIRR formatted files

The `airr` package contains the function `write_rearrangement` to write
Rearrangement records to the AIRR TSV format.

### Writing Rearrangements

```{r}
x1 <- file.path(tempdir(), "airr_out.tsv")
write_rearrangement(rearrangement, x1)
```

### Writing AIRR Data Models

AIRR Data Model records can be written to either YAML or JSON using the `write_airr` function.

```{r}
x2 <- file.path(tempdir(), "airr_repertoire_out.yaml")
write_airr(repertoire, x2, format="yaml")

x3 <- file.path(tempdir(), "airr_germline_out.json")
write_airr(germline, x3, format="json")
```

## Validating AIRR data structures

The `airr` package contains the function `validate_rearrangement` to validate
tabular (`data.frame`) Rearrangement records and AIRR Data Model objects, respectively.

```{r}
# Validate Rearrangement data.frame
validate_rearrangement(rearrangement)

# Validate an AIRR Data Model
validate_airr(repertoire)

# Validate AIRR Data Model records individual 
validate_airr(germline, each=TRUE)
```

## References

1. Breden, F., E. T. Luning Prak, B. Peters, F. Rubelt, C. A. Schramm, C. E. Busse, 
    J. A. Vander Heiden, et al. 2017. Reproducibility and Reuse of Adaptive Immune 
    Receptor Repertoire Data. *Front Immunol* 8: 1418.
2. Freeman, J. D., R. L. Warren, J. R. Webb, B. H. Nelson, and R. A. Holt. 2009. 
    Profiling the T-cell receptor beta-chain repertoire by massively parallel sequencing.
    *Genome Res* 19 (10): 1817-24.
3. Robins, H. S., P. V. Campregher, S. K. Srivastava, A. Wacher, C. J. Turtle, 
    O. Kahsai, S. R. Riddell, E. H. Warren, and C. S. Carlson. 2009. 
    Comprehensive assessment of T-cell receptor beta-chain diversity in alphabeta T cells.
    *Blood* 114 (19): 4099-4107.
4. Rubelt, F., C. E. Busse, S. A. C. Bukhari, J. P. Burckert, E. Mariotti-Ferrandiz, 
    L. G. Cowell, C. T. Watson, et al. 2017. Adaptive Immune Receptor Repertoire Community 
    recommendations for sharing immune-repertoire sequencing data.
    *Nat Immunol* 18 (12): 1274-8.
5. Weinstein, J. A., N. Jiang, R. A. White, D. S. Fisher, and S. R. Quake. 2009.
    High-throughput sequencing of the zebrafish antibody repertoire.
    *Science* 324 (5928): 807-10.

---
title: 'AIRR Data Representation Reference Library Usage'
author: "The AIRR Community"
date: '`r Sys.Date()`'
output:
  html_document:
    standalone: true
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    theme: readable
    toc: yes 
  md_document:
    fig_height: 4
    fig_width: 7.5
    preserve_yaml: no
    toc: yes
  pdf_document:
    dev: pdf
    fig_height: 4
    fig_width: 7.5
    highlight: pygments
    toc: yes
geometry: margin=1in
fontsize: 11pt
vignette: >
  %\VignetteIndexEntry{Usage Vignette}
  %\usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
---

## Introduction

Since the use of High-throughput sequencing (HTS) was first introduced to analyze 
immunoglobulin (B-cell receptor, antibody) and T-cell receptor repertoires 
(Freeman et al, 2009; Robins et al, 2009; Weinstein et al, 2009), 
the increasing number of studies making use of this technique has produced enormous 
amounts of data and there exists a pressing need to develop and adopt common standards, 
protocols, and policies for generating and sharing data sets. The [Adaptive Immune 
Receptor Repertoire (AIRR) Community](http://airr-community.org) formed in 
2015 to address this challenge (Breden et al, 2017) and has stablished the set of 
minimal metadata elements (MiAIRR) required for describing published AIRR 
datasets (Rubelt et al, 2017) as well as file formats to represent this data in a 
machine-readable form. The `airr` R package provide read, write and validation of data 
following the AIRR Data Representation schemas. This vignette provides a set of 
simple use examples.

### AIRR Data Standards

The AIRR Community's recommendations for a minimal set of metadata that should be 
used to describe an AIRR-seq data set when published or deposited in a 
AIRR-compliant public repository are described in Rubelt et al, 2017. The primary 
aim of this effort is to make published AIRR datasets FAIR (findable, accessible, 
interoperable, reusable); with sufficient detail such that a person skilled in the 
art of AIRR sequencing and data analysis will be able to reproduce the 
experiment and data analyses that were performed.

Following this principles, V(D)J reference alignment annotations are 
saved in standard tab-delimited files (TSV) with associated metadata 
provided in accompanying YAML formatted files. The column names 
and field names in these files have been defined by the 
AIRR Data Representation Working Group using a controlled vocabulary of 
standardized terms and types to refer to each piece of information.

## Reading AIRR formatted files

The `airr` package contains the function `read_rearrangement` to read and validate
files containing AIRR Rearrangement records, where a Rearrangement record describes 
the collection of optimal annotations on a single sequence that has undergone V(D)J reference alignment. 
The usage is straightforward, as the file format is a typical tabulated file. 
The argument that needs attention is `base`, with possible values `"0"` and `"1"`. 
`base` denotes the starting index for positional fields in the input file. 
Positional fields are those that contain alignment coordinates and names 
ending in "_start" and "_end". If the input file is using 1-based closed 
intervals (R style), as defined by the standard, then positional fields will 
not be modified under the default setting of `base="1"`. If the input file is using 
0-based coordinates with half-open intervals (python style), then 
positional fields may be converted to 1-based closed 
intervals using the argument `base="0"`.

### Reading Rearrangements
```{r}
# Imports
library(airr)
library(tibble)

# Read Rearrangement example file
f1 <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")
rearrangement <- read_rearrangement(f1)
glimpse(rearrangement)
```

### Reading AIRR Data Models

AIRR Data Model records, such as Repertoire and GermlineSet, can be read from either a YAML 
or JSON formatted file into a nested list.

```{r}
# Read Repertoire example file
f2 <- system.file("extdata", "repertoire-example.yaml", package="airr")
repertoire <- read_airr(f2)
glimpse(repertoire)

# Read GermlineSet example file
f3 <- system.file("extdata", "germline-example.json", package="airr")
germline <- read_airr(f3)
glimpse(germline)
```


## Writing AIRR formatted files

The `airr` package contains the function `write_rearrangement` to write
Rearrangement records to the AIRR TSV format.

### Writing Rearrangements

```{r}
x1 <- file.path(tempdir(), "airr_out.tsv")
write_rearrangement(rearrangement, x1)
```

### Writing AIRR Data Models

AIRR Data Model records can be written to either YAML or JSON using the `write_airr` function.

```{r}
x2 <- file.path(tempdir(), "airr_repertoire_out.yaml")
write_airr(repertoire, x2, format="yaml")

x3 <- file.path(tempdir(), "airr_germline_out.json")
write_airr(germline, x3, format="json")
```

## Validating AIRR data structures

The `airr` package contains the function `validate_rearrangement` to validate
tabular (`data.frame`) Rearrangement records and AIRR Data Model objects, respectively.

```{r}
# Validate Rearrangement data.frame
validate_rearrangement(rearrangement)

# Validate an AIRR Data Model
validate_airr(repertoire)

# Validate AIRR Data Model records individual 
validate_airr(germline, each=TRUE)
```

## References

1. Breden, F., E. T. Luning Prak, B. Peters, F. Rubelt, C. A. Schramm, C. E. Busse, 
    J. A. Vander Heiden, et al. 2017. Reproducibility and Reuse of Adaptive Immune 
    Receptor Repertoire Data. *Front Immunol* 8: 1418.
2. Freeman, J. D., R. L. Warren, J. R. Webb, B. H. Nelson, and R. A. Holt. 2009. 
    Profiling the T-cell receptor beta-chain repertoire by massively parallel sequencing.
    *Genome Res* 19 (10): 1817-24.
3. Robins, H. S., P. V. Campregher, S. K. Srivastava, A. Wacher, C. J. Turtle, 
    O. Kahsai, S. R. Riddell, E. H. Warren, and C. S. Carlson. 2009. 
    Comprehensive assessment of T-cell receptor beta-chain diversity in alphabeta T cells.
    *Blood* 114 (19): 4099-4107.
4. Rubelt, F., C. E. Busse, S. A. C. Bukhari, J. P. Burckert, E. Mariotti-Ferrandiz, 
    L. G. Cowell, C. T. Watson, et al. 2017. Adaptive Immune Receptor Repertoire Community 
    recommendations for sharing immune-repertoire sequencing data.
    *Nat Immunol* 18 (12): 1274-8.
5. Weinstein, J. A., N. Jiang, R. A. White, D. S. Fisher, and S. R. Quake. 2009.
    High-throughput sequencing of the zebrafish antibody repertoire.
    *Science* 324 (5928): 807-10.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Interface.R
\name{read_airr}
\alias{read_airr}
\title{Read an AIRR Data Model file in YAML or JSON format}
\usage{
read_airr(
  file,
  format = c("auto", "yaml", "json"),
  validate = TRUE,
  model = TRUE
)
}
\arguments{
\item{file}{path to the input file.}

\item{format}{format of the input file. Must be one of \code{"auto"}, \code{"yaml"}, or 
\code{"json"}. If \code{"auto"} (default), the format will be 
detected from the \code{file} extension.}

\item{validate}{run schema validation if \code{TRUE}.}

\item{model}{if \code{TRUE} validate only AIRR DataFile defined objects. If \code{FALSE} 
attempt validation of all objects in \code{data}.
Ignored if \code{validate=FALSE}}
}
\value{
A named nested \code{list} contained in the AIRR Data Model with the top-level
          names reflecting the individual AIRR objects.
}
\description{
\code{read_airr} loads a YAML or JSON file containing AIRR Data Model records.
}
\examples{
# Get path to the Reportoire and GermlineSet example files
f1 <- system.file("extdata", "repertoire-example.yaml", package="airr")
f2 <- system.file("extdata", "germline-example.json", package="airr")

# Load data files
repertoire <- read_airr(f1)
germline <- read_airr(f2)

}
\seealso{
See \link{Schema} for the AIRR schema definition objects.
See \link{write_airr} for writing AIRR Data Model records in YAML or JSON format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Interface.R
\name{validate_tabular}
\alias{validate_tabular}
\alias{validate_rearrangement}
\title{Validate tabular AIRR data}
\usage{
validate_tabular(data, schema)

validate_rearrangement(data)
}
\arguments{
\item{data}{\code{data.frame} of tabular data to validate.}

\item{schema}{\code{Schema} object defining the data standard of the table.}
}
\value{
Returns \code{TRUE} if the input \code{data} is compliant and
          \code{FALSE} if not.
}
\description{
\code{validate_tabular} validates compliance of the contents of a \code{data.frame} 
to the AIRR standards.
}
\details{
\code{validate_rearrangement} validates the standards compliance of AIRR Rearrangement 
data stored in a \code{data.frame}
}
\examples{
# Get path to the rearrangement-example file
file <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")

# Load data file
df <- read_rearrangement(file)

# Validate a data.frame against the Rearrangement schema
validate_rearrangement(df)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Interface.R
\name{write_tabular}
\alias{write_tabular}
\alias{write_rearrangement}
\alias{write_alignment}
\title{Write an AIRR tabular data}
\usage{
write_tabular(data, file, schema, base = c("1", "0"), ...)

write_rearrangement(data, file, base = c("1", "0"), ...)

write_alignment(data, file, base = c("1", "0"), ...)
}
\arguments{
\item{data}{\code{data.frame} of Rearrangement data.}

\item{file}{output file name.}

\item{schema}{\code{Schema} object defining the output format.}

\item{base}{starting index for positional fields in the output file.
Fields in the input \code{data} are assumed to be 1-based
closed-intervals (R style).
If \code{"1"}, then these fields will not be modified.
If \code{"0"}, then fields ending in \code{_start} and \code{_end}
will be converted to 0-based half-open intervals (python style)
in the output file.}

\item{...}{additional arguments to pass to \link[readr]{write_delim}.}
}
\description{
\code{write_tabular} writes a TSV containing AIRR tabular records.
}
\details{
\code{write_rearrangement} writes a \code{data.frame} containing AIRR Rearrangement data to TSV.

\code{write_alignment} writes a \code{data.frame} containing AIRR Alignment data to TSV.
}
\examples{
# Get path to the rearrangement-example file
file <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")

# Load data file
df <- read_rearrangement(file)

# Write a Rearrangement data file
outfile <- file.path(tempdir(), "output.tsv")
write_tabular(df, outfile, schema=RearrangementSchema)

}
\seealso{
See \link{Schema} for the AIRR schema object definition.
See \link{read_tabular} for reading to AIRR files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Schema.R
\name{load_schema}
\alias{load_schema}
\title{Load a schema definition}
\usage{
load_schema(definition)
}
\arguments{
\item{definition}{name of the schema definition.}
}
\value{
A \link{Schema} object for the \code{definition}.
}
\description{
\code{load_schema} loads an AIRR object definition from the internal
definition set.
}
\details{
Valid definitions include:
\itemize{
  \item   \code{"Rearrangement"}
  \item   \code{"Alignment"}
  \item   \code{"Repertoire"}
  \item   \code{"Study"}
  \item   \code{"Subject"}
  \item   \code{"Diagnosis"}
  \item   \code{"Sample"}
  \item   \code{"SampleProcessing"}
  \item   \code{"DataProcessing"}
  \item   \code{"GermlineSet"}
  \item   \code{"GenotypeSet"}
}
}
\examples{
# Load the Rearrangement definition
schema <- load_schema("Rearrangement")

# Load the Repertoire definition
schema <- load_schema("Repertoire")

}
\seealso{
See \link{Schema} for the return object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Interface.R
\name{write_airr}
\alias{write_airr}
\title{Write AIRR Data Model records to YAML or JSON files}
\usage{
write_airr(
  data,
  file,
  format = c("auto", "yaml", "json"),
  validate = TRUE,
  model = TRUE
)
}
\arguments{
\item{data}{\code{list} containing AIRR Model Records.}

\item{file}{output file name.}

\item{format}{format of the output file. Must be one of \code{"auto"}, \code{"yaml"}, or 
\code{"json"}. If \code{"auto"} (default), the format will be 
detected from the \code{file} extension.}

\item{validate}{run schema validation prior to write if \code{TRUE}.}

\item{model}{if \code{TRUE} validate and write only AIRR DataFile defined objects. 
If \code{FALSE} attempt validation and write of all objects in \code{data}.}
}
\description{
\code{write_airr} writes a YAML or JSON file containing AIRR Data Model records.
}
\examples{
# Get path to the repertoire-example file
file <- system.file("extdata", "repertoire-example.yaml", package="airr")

# Load data file
repertoire <- read_airr(file)

# Write a Rearrangement data file
outfile <- file.path(tempdir(), "output.yaml")
write_airr(repertoire, outfile)

}
\seealso{
See \link{Schema} for the AIRR schema definition objects.
See \link{read_airr} for reading to AIRR Data Model files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Schema.R
\docType{class}
\name{Schema-class}
\alias{Schema-class}
\alias{Schema}
\alias{names,Schema-method}
\alias{Schema-method}
\alias{[,Schema,character-method}
\alias{$,Schema-method}
\alias{InfoSchema}
\alias{DataFileSchema}
\alias{AlignmentSchema}
\alias{RearrangementSchema}
\alias{RepertoireSchema}
\alias{GermlineSetSchema}
\alias{GenotypeSetSchema}
\alias{AIRRSchema}
\title{S4 class defining an AIRR standard schema}
\format{
A \code{Schema} object.

An object of class \code{Schema} of length 1.

An object of class \code{Schema} of length 1.

An object of class \code{Schema} of length 1.

An object of class \code{Schema} of length 1.

An object of class \code{Schema} of length 1.

An object of class \code{Schema} of length 1.

An object of class \code{Schema} of length 1.

An object of class \code{list} of length 26.
}
\usage{
\S4method{names}{Schema}(x)

\S4method{[}{Schema,character}(x, i)

\S4method{$}{Schema}(x, name)

InfoSchema

DataFileSchema

AlignmentSchema

RearrangementSchema

RepertoireSchema

GermlineSetSchema

GenotypeSetSchema

AIRRSchema
}
\arguments{
\item{x}{\code{Schema} object.}

\item{i}{field name.}

\item{name}{field name.}
}
\description{
\code{Schema} defines a common data structure for AIRR Data Representation 
standards.
}
\details{
The following predefined Schema objects are defined:

\code{InfoSchema}: AIRR Info \code{Schema}.

\code{DataFileSchema}: AIRR DataFile \code{Schema}.

\code{AlignmentSchema}: AIRR Alignment \code{Schema}.

\code{RearrangementSchema}: AIRR Rearrangement \code{Schema}.

\code{RepertoireSchema}: AIRR Repertoire \code{Schema}.

\code{GermlineSetSchema}: AIRR GermlineSet \code{Schema}.

\code{GenotypeSetSchema}: AIRR GenotypeSet \code{Schema}.

\code{AIRRSchema}: named list containing all non-experimental 
                              AIRR \code{Schema} objects.
}
\section{Slots}{

\describe{
\item{\code{definition}}{name of the schema definition.}

\item{\code{required}}{\code{character} vector of required fields.}

\item{\code{optional}}{\code{character} vector of non-required fields.}

\item{\code{properties}}{\code{list} of field definitions.}

\item{\code{info}}{\code{list} schema information.}
}}

\seealso{
See \link{load_schema} for loading a \code{Schema} from the definition set.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Schema.R
\name{ExampleData}
\alias{ExampleData}
\title{Example AIRR data}
\format{
\itemize{
  \item  \code{extdata/rearrangement-example.tsv.gz}: Rearrangement TSV file.
  \item  \code{extdata/repertoire-example.yaml}: Repertoire YAML file.
  \item  \code{extdata/germline-example.json}: GermlineSet and GenotypeSet JSON file.
}
}
\description{
Example data files compliant with the the AIRR Data Representation standards.
}
\examples{
# Load Rearrangement example
file <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")
rearrangement <- read_rearrangement(file)

# Load Repertoire example
file <- system.file("extdata", "repertoire-example.yaml", package="airr")
repertoire <- read_airr(file)

# Load GermlineSet and GenotypeSet examples
file <- system.file("extdata", "germline-example.json", package="airr")
germline <- read_airr(file)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Interface.R
\name{validate_airr}
\alias{validate_airr}
\title{Validate an AIRR Data Model nested list representation}
\usage{
validate_airr(data, model = TRUE, each = FALSE)
}
\arguments{
\item{data}{\code{list} containing records of an AIRR Data Model objected imported from 
a YAML or JSON representation.}

\item{model}{if \code{TRUE} validate only AIRR DataFile defined objects. If \code{FALSE} 
attempt validation of all objects in \code{data}.}

\item{each}{if \code{TRUE} return a logical vector with results for each object in \code{data}
instead of a single \code{TRUE} or \code{FALSE} value.}
}
\value{
Returns \code{TRUE} if the input \code{data} is compliant with AIRR standards and
          \code{FALSE} if not. If \code{each=TRUE} is set, then a vector with results for each
          each object in \code{data} is returned instead.
}
\description{
\code{validate_airr} validates the fields in a named nested list representation of the 
AIRR Data Model. Typically, generating by reading of JSON or YAML formatted AIRR files.
}
\examples{
# Get path to the rearrangement-example file
f1 <- system.file("extdata", "repertoire-example.yaml", package="airr")
f2 <- system.file("extdata", "germline-example.json", package="airr")

# Load data file
repertoire <- read_airr(f1)
germline <- read_airr(f2)

# Validate a single record
validate_airr(repertoire)

# Return validation for individual objects
validate_airr(germline, each=TRUE)

}
\seealso{
See \link{Schema} for the AIRR schema definitions.
See \link{read_airr} for loading AIRR Data Models from a file.
See \link{write_airr} for writing AIRR Data Models to a file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Interface.R
\name{read_tabular}
\alias{read_tabular}
\alias{read_rearrangement}
\alias{read_alignment}
\title{Read AIRR tabular data}
\usage{
read_tabular(file, schema, base = c("1", "0"), aux_types = NULL, ...)

read_rearrangement(file, base = c("1", "0"), ...)

read_alignment(file, base = c("1", "0"), ...)
}
\arguments{
\item{file}{input file path.}

\item{schema}{\code{Schema} object defining the output format.}

\item{base}{starting index for positional fields in the input file.
If \code{"1"}, then these fields will not be modified.
If \code{"0"}, then fields ending in \code{"_start"} and \code{"_end"}
are 0-based half-open intervals (python style) in the input file
and will be converted to 1-based closed-intervals (R style).}

\item{aux_types}{named vector or list giving the type for fields that are not
defined in \code{schema}. The field name is the name, the value
the type, denoted by one of \code{"c"} (character), \code{"l"} (logical),
\code{"i"} (integer), \code{"d"} (double), or \code{"n"} (numeric).}

\item{...}{additional arguments to pass to \link[readr]{read_delim}.}
}
\value{
A \code{data.frame} of the TSV file with appropriate type and position conversion
          for fields defined in the specification.
}
\description{
\code{read_tabular} reads a tab-delimited (TSV) file containing tabular AIRR records.
}
\details{
\code{read_rearrangement} reads an AIRR TSV containing Rearrangement data.

\code{read_alignment} reads an AIRR TSV containing Alignment data.
}
\examples{
# Get path to the rearrangement-example file
file <- system.file("extdata", "rearrangement-example.tsv.gz", package="airr")

# Load data file
df <- read_rearrangement(file)

}
\seealso{
See \link{Schema} for the AIRR schema object definition.
See \link{write_tabular} for writing AIRR data.
}
